"use strict";
var Busboy = require("busboy");
var Promise = require("bluebird");
var definitions_1 = require("./definitions");
var internal_1 = require("./internal");
var ResponseObject_1 = require("./ResponseObject");
var ExpressAdapter = (function () {
    function ExpressAdapter(app, services) {
        this.app = app;
        this.services = services;
    }
    ExpressAdapter.prototype.addMiddleware = function (configuration, path, controller, handlerName, handler) {
        var _this = this;
        if (path) {
            this.app.use(path, function (req, res, next) {
                return internal_1.callRequestHandler(_this, handler, controller, configuration, handlerName, { req: req, res: res }, next);
            });
        }
        else {
            this.app.use(function (req, res, next) {
                return internal_1.callRequestHandler(_this, handler, controller, configuration, handlerName, { req: req, res: res }, next);
            });
        }
    };
    ExpressAdapter.prototype.addRoute = function (configuration, method, path, controller, handlerName, handler, middlewares) {
        var _this = this;
        var routeFunc = function (req, res, next) {
            return internal_1.callRequestHandler(_this, handler, controller, configuration, handlerName, { req: req, res: res }, next);
        };
        var args = [path];
        if (middlewares) {
            middlewares.forEach(function (middleware) { return args.push(middleware); });
        }
        args.push(routeFunc);
        this.app[method].apply(this.app, args);
    };
    ExpressAdapter.prototype.getParameterWithConfig = function (paramConfig, _a) {
        var req = _a.req;
        switch (paramConfig.type) {
            case definitions_1.ParameterType.PATH_PARAMETER:
                return this.transformParamConfig(paramConfig, req.params[paramConfig.name]);
            case definitions_1.ParameterType.QUERY_PARAMETER:
                return this.transformParamConfig(paramConfig, req.query[paramConfig.name]);
            case definitions_1.ParameterType.BODY_PARAMETER:
                if (paramConfig.name) {
                    return this.transformParamConfig(paramConfig, req.body[paramConfig.name]);
                }
                else {
                    return req.body;
                }
            case definitions_1.ParameterType.MULTIPART_PARAMETER: {
                return new Promise(function (resolve, reject) {
                    var busboy = new Busboy({
                        headers: req.headers,
                    });
                    var files = [];
                    var fields = {};
                    busboy.on("file", function (fieldname, file, filename, encoding, mimetype) {
                        var filedata = [];
                        file.on("data", function (data) { return filedata.push(data); });
                        file.on("end", function () {
                            var fileInfo = {
                                data: Buffer.concat(filedata),
                                name: filename,
                                encoding: encoding,
                                mimetype: mimetype,
                                fieldname: fieldname,
                            };
                            files.push(fileInfo);
                        });
                    });
                    busboy.on("field", function (fieldname, value) { return fields[fieldname] = value; });
                    busboy.on("finish", function () {
                        resolve({
                            files: files,
                            fields: fields,
                        });
                    });
                    busboy.on("error", reject);
                    req.pipe(busboy);
                });
            }
            case definitions_1.ParameterType.COOKIE_PARAMETER:
                return this.transformParamConfig(paramConfig, req.cookies[paramConfig.name]);
            case definitions_1.ParameterType.HEADER_PARAMETER:
                return this.transformParamConfig(paramConfig, req.headers[paramConfig.name.toLowerCase()]);
            case definitions_1.ParameterType.SERVICE_PARAMETER: {
                if (paramConfig.name) {
                    return this.services.indexOf(paramConfig.name) === -1 ? undefined : req[paramConfig.name];
                }
                else {
                    var serv_1 = {};
                    this.services.forEach(function (name) { return serv_1[name] = req[name]; });
                    return serv_1;
                }
            }
            default:
                throw new Error("Unknown decorator parameter type");
        }
    };
    ExpressAdapter.prototype.send = function (statusCode, data, _a, headers) {
        var res = _a.res;
        if (data instanceof ResponseObject_1.ResponseObject) {
            res.status(statusCode);
            data.send(res);
            return;
        }
        if (headers) {
            headers.forEach(function (header) { return res.set(header.field, header.value); });
        }
        if (typeof data === "number") {
            data = data.toString();
        }
        res.status(statusCode).send(data);
    };
    ExpressAdapter.prototype.sendJson = function (statusCode, data, _a, headers) {
        var res = _a.res;
        if (headers) {
            headers.forEach(function (header) { return res.set(header.field, header.value); });
        }
        res.status(statusCode).json(data);
    };
    ExpressAdapter.prototype.transformParamConfig = function (paramConfig, value) {
        return this.transformParam(paramConfig.name, value, paramConfig.valueType, paramConfig.required);
    };
    ExpressAdapter.prototype.makeParamError = function (error) {
        error.statusCode = error.status = error.code = 422;
        return error;
    };
    ExpressAdapter.prototype.transformParam = function (name, value, type, required) {
        if (value === undefined || value === null) {
            if (!required) {
                return value;
            }
            var typeName = type === undefined ? "any" : type.name;
            throw this.makeParamError(new Error("Required " + typeName + " parameter '" + name + "' missing"));
        }
        if (type === undefined) {
            return value;
        }
        if (type === String) {
            return value + "";
        }
        else if (type === Number) {
            if (typeof value === "number") {
                return value;
            }
            else if (typeof value === "string") {
                var val = Number(value);
                if (!isNaN(val)) {
                    return val;
                }
            }
            else if (typeof value === "object" && value instanceof Number) {
                return value.valueOf();
            }
            throw this.makeParamError(new Error("Could not transform value for '" + name + "' to type number"));
        }
        else if (type === Boolean) {
            if (typeof value === "number") {
                if (value === 1) {
                    return true;
                }
                else if (value === 0) {
                    return false;
                }
            }
            else if (typeof value === "boolean") {
                return value;
            }
            if (typeof value === "string") {
                if (value.toLowerCase() === "true" || value.toLowerCase() === "yes" || value === "1") {
                    return true;
                }
                else if (value.toLowerCase() === "false" || value.toLowerCase() === "no" || value === "0") {
                    return false;
                }
            }
            else if (typeof value === "object" && value instanceof Boolean) {
                return value.valueOf();
            }
            throw this.makeParamError(new Error("Could not transform value for '" + name + "' to type boolean"));
        }
        else if (type === Date) {
            if (typeof value === "string") {
                var val = new Date(value);
                if (!isNaN(val.valueOf())) {
                    return val;
                }
            }
            else if (typeof value === "number") {
                var val = new Date(value);
                if (!isNaN(val.valueOf())) {
                    return val;
                }
            }
            else if (typeof value === "object" && value instanceof Date) {
                return value;
            }
            throw this.makeParamError(new Error("Could not transform value for '" + name + "' to type Date"));
        }
        else if (typeof type === "function") {
            if (value instanceof type) {
                return value;
            }
            throw this.makeParamError(new Error("Could not transform value for '" + name + "' to type '" + type.name + "'"));
        }
        throw new Error("Could not perform type-transform for unknown type " + type.name);
    };
    return ExpressAdapter;
}());
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = ExpressAdapter;
//# sourceMappingURL=ExpressAdapter.js.map