import * as express from "express";
import * as Busboy from "busboy";
import * as Promise from "bluebird";

import {IAdapterData, IControllerConfiguration, ParameterType, IParameterConfiguration, IHeader} from "./definitions";
import {callRequestHandler} from "./internal";
import {ResponseObject} from "./ResponseObject";

/**
 * Contains all express specific code.
 */
export default class ExpressAdapter {
    constructor(protected app: express.Express, protected services: string[]) {
    }

    /**
     * Adds a middle function to the app.
     */
    public addMiddleware(configuration: IControllerConfiguration, path: string,
        controller: any, handlerName: string, handler: Function): void {
        if (path) {
            this.app.use(path, (req: express.Request, res: express.Response, next: Function) =>
                callRequestHandler(this, handler, controller, configuration, handlerName, { req, res }, next)
            );
        } else {
            this.app.use((req: express.Request, res: express.Response, next: Function) =>
                callRequestHandler(this, handler, controller, configuration, handlerName, { req, res }, next)
            );
        }
    }

    /**
     * Adds a route function to the app.
     */
    public addRoute(configuration: IControllerConfiguration, method: string, path: string,
        controller: any, handlerName: string, handler: Function, middlewares: express.RequestHandler[]): void {

        // The routing function
        let routeFunc = (req: express.Request, res: express.Response, next: Function) =>
            callRequestHandler(this, handler, controller, configuration, handlerName, { req, res }, next);

        // An argument array with the path, middlewares and route function
        let args: any[] = [path];
        if (middlewares) {
            middlewares.forEach((middleware: express.RequestHandler) => args.push(middleware));
        }
        args.push(routeFunc);

        // Tell Express the new handler
        this.app[method].apply(this.app, args);
    }

    public getParameterWithConfig(paramConfig: IParameterConfiguration, {req}: IAdapterData): any {
        switch (paramConfig.type) {
            case ParameterType.PATH_PARAMETER:
                return this.transformParamConfig(paramConfig, req.params[paramConfig.name]);
            case ParameterType.QUERY_PARAMETER:
                return this.transformParamConfig(paramConfig, req.query[paramConfig.name]);
            case ParameterType.BODY_PARAMETER:
                if (paramConfig.name) {
                    return this.transformParamConfig(paramConfig, req.body[paramConfig.name]);
                } else {
                    return req.body;
                }
            case ParameterType.MULTIPART_PARAMETER: {
                return new Promise((resolve: Function, reject: Function) => {
                    let busboy = new Busboy({
                        headers: req.headers,
                    });

                    let files: Object[] = [];
                    let fields: Object = {};

                    busboy.on("file", (fieldname: string, file: any, filename: string, encoding: string, mimetype: string) => {
                        let filedata: Buffer[] = [];

                        file.on("data", (data: Buffer) => filedata.push(data));

                        file.on("end", () => {
                            let fileInfo = {
                                data: Buffer.concat(filedata),
                                name: filename,
                                encoding: encoding,
                                mimetype: mimetype,
                                fieldname: fieldname,
                            };

                            files.push(fileInfo);
                        });
                    });

                    busboy.on("field", (fieldname: string, value: any) => fields[fieldname] = value);

                    busboy.on("finish", () => {
                        resolve({
                            files: files,
                            fields: fields,
                        });
                    });

                    busboy.on("error", reject);

                    req.pipe(busboy);
                });
            }
            case ParameterType.COOKIE_PARAMETER:
                return this.transformParamConfig(paramConfig, req.cookies[paramConfig.name]);
            case ParameterType.HEADER_PARAMETER:
                return this.transformParamConfig(paramConfig, req.headers[paramConfig.name.toLowerCase()]);
            case ParameterType.SERVICE_PARAMETER: {
                if (paramConfig.name) {
                    return this.services.indexOf(paramConfig.name) === -1 ? undefined : req[paramConfig.name];
                } else {
                    let serv: Object = {};
                    this.services.forEach((name: string) => serv[name] = req[name]);
                    return serv;
                }
            }
            default:
                throw new Error("Unknown decorator parameter type");
        }
    }

    public send(statusCode: number, data: any, {res}: IAdapterData, headers?: IHeader[]): void {
        if (data instanceof ResponseObject) {
            res.status(statusCode);
            data.send(res);
            return;
        }

        if (headers) {
            headers.forEach((header: IHeader) => res.set(header.field, header.value));
        }

        if (typeof data === "number") {
            data = data.toString();
        }

        res.status(statusCode).send(data);
    }

    public sendJson(statusCode: number, data: any, {res}: IAdapterData, headers?: IHeader[]): void {
        if (headers) {
            headers.forEach((header: IHeader) => res.set(header.field, header.value));
        }
        res.status(statusCode).json(data);
    }

    private transformParamConfig(paramConfig: IParameterConfiguration, value: any): any {
        return this.transformParam(paramConfig.name, value, paramConfig.valueType, paramConfig.required);
    }

    private makeParamError(error: Error): Error {
        (<any>error).statusCode = (<any>error).status = (<any>error).code = 422;
        return error;
    }

    private transformParam(name: string, value: any, type?: Function, required?: boolean): any {
        if (value === undefined || value === null) {
            if (!required) {
                return value;
            }

            const typeName = type === undefined ? "any" : (<any>type).name;
            throw this.makeParamError(new Error(`Required ${typeName} parameter '${name}' missing`));
        }

        if (type === undefined) {
            return value;
        }

        if (type === String) {
            return value + "";
        } else if (type === Number) {
            if (typeof value === "number") {
                return value;
            } else if (typeof value === "string") {
                const val = Number(value);
                if (!isNaN(val)) {
                    return val;
                }
            } else if (typeof value === "object" && value instanceof Number) {
                return value.valueOf();
            }

            throw this.makeParamError(new Error(`Could not transform value for '${name}' to type number`));
        } else if (type === Boolean) {
            if (typeof value === "number") {
                if (value === 1) {
                    return true;
                } else if (value === 0) {
                    return false;
                }
            } else if (typeof value === "boolean") {
                return value;
            } if (typeof value === "string") {
                if (value.toLowerCase() === "true" || value.toLowerCase() === "yes" || value === "1") {
                    return true;
                } else if (value.toLowerCase() === "false" || value.toLowerCase() === "no" || value === "0") {
                    return false;
                }
            } else if (typeof value === "object" && value instanceof Boolean) {
                return value.valueOf();
            }

            throw this.makeParamError(new Error(`Could not transform value for '${name}' to type boolean`));
        } else if (type === Date) {
            if (typeof value === "string") {
                const val = new Date(value);
                if (!isNaN(val.valueOf())) {
                    return val;
                }
            } else if (typeof value === "number") {
                const val = new Date(value);
                if (!isNaN(val.valueOf())) {
                    return val;
                }
            } else if (typeof value === "object" && value instanceof Date) {
                return value;
            }

            throw this.makeParamError(new Error(`Could not transform value for '${name}' to type Date`));
        } else if (typeof type === "function") {
            if (value instanceof type) {
                return value;
            }

            throw this.makeParamError(new Error(`Could not transform value for '${name}' to type '${(<any>type).name}'`));
        }

        throw new Error(`Could not perform type-transform for unknown type ${(<any>type).name}`);
    }
}
