/**
 * Logging decorator.
 *
 * On a class: logs when a new instance is created.
 * On a property: logs sets and gets.
 * On a method: logs when the method is called.
 *
 * Add @log before a method to make it log.
 */
export function Log(...args: any[]): any {
    switch (args.length) {
        case 1:
            return logClass.apply(this, args);
        case 2:
            return logProperty.apply(this, args);
        case 3:
            if (typeof args[2] === "number") {
                return logParameter.apply(this, args);
            }
            return logMethod.apply(this, args);
        default:
            throw new Error("Decorators are not valid here.");
    }
}

function logMethod(targetClass: any, methodName: string, descriptor: any): any {
    let originalMethod: Function = descriptor.value;

    descriptor.value = function(...args: any[]): any {

        let metadataKey = `log_${methodName}_parameters`;
        let indices: number[] = targetClass[metadataKey];

        if (Array.isArray(indices)) {
            for (let i = 0; i < args.length; ++i) {
                if (indices.indexOf(i) !== -1) {
                    let arg = args[i];
                    let argStr = JSON.stringify(arg) || arg.toString();

                    console.log(`${methodName} arg[${i}]: ${argStr}`);
                }
            }

            let result: any = originalMethod.apply(this, args);

            // Make the function return the result
            return result;
        } else {
            // Make a string of the parameters for display
            let a: string = args.map((arg: any) => {
                try {
                    return JSON.stringify(arg);
                } catch (err) {
                    return "!!ERROR: " + err.message + "!!";
                }
            }).join();

            // Execute the function
            let result = originalMethod.apply(this, args);

            // Stringify the result
            let r = JSON.stringify(result);

            // Show arguments and result
            console.log(`Call: ${methodName}(${a}) => ${r}`);

            // Make the function return the result
            return result;
        }
    };

    return descriptor;
}

function logParameter(targetClass: any, methodName: string, index: number): void {
    let metadataKey = `log_${methodName}_parameters`;

    if (Array.isArray(targetClass[metadataKey])) {
        targetClass[metadataKey].push(index);
    } else {
        targetClass[metadataKey] = [index];
    }
}

interface IFace { }
function logClass(targetClass: { name?: string, new (...args: any[]): IFace }): void {
    let originalClass = targetClass;

    return <any>class extends targetClass {
        constructor(...args: any[]) {
            console.log(`new: ${originalClass.name}`);
            super(...args);
        }
    };
}

function logProperty(targetClass: any, propertyName: string): void {
    let value: any = this[propertyName];

    let getter = function(): any {
        console.log(`Get: ${propertyName} => ${value}`);
        return value;
    };

    let setter = function(newValue: any): void {
        console.log(`Set: ${propertyName} <= ${newValue} (was ${value})`);
        value = newValue;
    };

    // Recreate the property
    if (delete this[propertyName]) {
        Object.defineProperty(targetClass, propertyName, {
            configurable: true,
            enumerable: true,
            get: getter,
            set: setter,
        });
    }
}
