import * as Promise from "bluebird";
import * as domain from "domain";
import * as express from "express";

import {IObjectWithControllerConfiguration, IParameterConfiguration, IControllerClass} from "./definitions";
import {IControllerConfiguration, SendType, ParameterType, IRoute, IMiddleware, IHeader} from "./definitions";

import ExpressAdapter from "./ExpressAdapter";

/**
 * Initialize a class for decoration.
 *
 * @param {IObjectWithControllerConfiguration} target the class.
 */
function addClassConfiguration(target: IObjectWithControllerConfiguration): void {
    // If the configuration does not exist yet, set a bunch of standard values.
    if (!target.$$controllerConfiguration) {
        target.$$controllerConfiguration = {
            methodsParameters: {},
            middlewares: [],
            root: null,
            routes: [],
            routesMiddlewares: {},
            sendTypes: {},
        };
    }
}

/**
 * Add a method decorator to given target method.
 *
 * Initializes the target if neccessary.
 *
 * @param {any}    target                 [description]
 * @param {string} methodName             [description]
 * @param {any}    parameterConfiguration [description]
 */
function addMethodDecorator(target: any, methodName: string, parameterConfiguration: IParameterConfiguration): void {
    // Prepare the class for route decorators
    addClassConfiguration(target);

    // If there are no method decorators for this method yet, create an empty list
    if (!target.$$controllerConfiguration.methodsParameters[methodName]) {
        target.$$controllerConfiguration.methodsParameters[methodName] = [];
    }

    // Add the configuration
    target.$$controllerConfiguration.methodsParameters[methodName].push(parameterConfiguration);
}

export function addRouteMiddleware(target: IObjectWithControllerConfiguration, methodName: string,
    middleware: express.RequestHandler): void {
    addClassConfiguration(target);

    if (!target.$$controllerConfiguration.routesMiddlewares[methodName]) {
        target.$$controllerConfiguration.routesMiddlewares[methodName] = [];
    }

    target.$$controllerConfiguration.routesMiddlewares[methodName].push(middleware);
}


/**
 * The controller decorator.
 *
 * @param  { string } root Root of the router the controller is at.
 */
export function ControllerDec(root: string): ClassDecorator {
    return function(target: IControllerClass): IControllerClass {
        // Initialize the configuration of the class
        addClassConfiguration(target.prototype);

        // Set the root path
        target.prototype.$$controllerConfiguration.root = root;

        return target;
    };
}

/**
 * The JSON decorator.
 */
export function SendJsonDec(target: IObjectWithControllerConfiguration, methodName: string,
    descriptor: TypedPropertyDescriptor<any>): void {
    // Initialize the configuration of the class
    addClassConfiguration(target);

    target.$$controllerConfiguration.sendTypes[methodName] = SendType.JSON;
}

/////////////////////////////////////////////////////////////////////
/// Factory functions
/////////////////////////////////////////////////////////////////////

/**
 * Generate a parameter decorator for given parameter type.
 *
 * @param {ParameterType} paramType [description]
 */
export function parameterDecoratorFactory(paramType: ParameterType): () => ParameterDecorator {
    return function(): ParameterDecorator {
        return function(_target: Function, methodName: string, index: number): void {
            const target: IObjectWithControllerConfiguration = <IObjectWithControllerConfiguration><any>_target;

            let paramConfig: IParameterConfiguration = { index, type: paramType };

            addMethodDecorator(target, methodName, paramConfig);
        };
    };
}

/**
 * Generate a parameter decorator for given parameter type. The decorator will have a name argument.
 *
 * @param {string} paramType parameter type
 */
export function parameterDecoratorWithNameFactory(paramType: ParameterType): (_?: string) => ParameterDecorator {
    return function(name?: string, type?: Function, required?: boolean): ParameterDecorator {
        return function(_target: Function, methodName: string, index: number): void {
            const target: IObjectWithControllerConfiguration = <IObjectWithControllerConfiguration><any>_target;

            let paramConfig: IParameterConfiguration = {
                index: index,
                name: name,
                type: paramType,
                valueType: type,
                required: required,
            };

            addMethodDecorator(target, methodName, paramConfig);
        };
    };
}

/**
 * Generate an HTTP method decorator for given HTTP method.
 *
 * @param {string} method The HTTP method.
 */
export function methodDecoratorFactory(method: string): (p?: string) => MethodDecorator {
    return function(path?: string): MethodDecorator {
        return function(target: IObjectWithControllerConfiguration, handlerName: string,
            descriptor: TypedPropertyDescriptor<Function>): void {
            // Make sure the class can be decorated
            addClassConfiguration(target);

            let route: IRoute = { method, path, handlerName };

            // Add a new rout
            target.$$controllerConfiguration.routes.push(route);
        };
    };
}

/////////////////////////////////////////////////////////////////////
/// Runtime functions
/////////////////////////////////////////////////////////////////////

export function callRequestHandler(adapter: ExpressAdapter, handler: Function, controller: any,
    configuration: IControllerConfiguration, handlerName: string, adapterRequestData: any, next: Function): void {
    // Create the list of parameters
    const paramList: any[] = [];
    if (configuration.methodsParameters[handlerName]) {
        configuration.methodsParameters[handlerName].forEach((paramConfig: IParameterConfiguration) => {
            paramList[paramConfig.index] = adapter.getParameterWithConfig(paramConfig, adapterRequestData);
        });
    }
    paramList.push(next);

    // Run the class method in a seperate domain.
    const dom = domain.create();

    dom.on("error", next);

    dom.run(() => {
        const result: any = handler.apply(controller, paramList);

        // Handle the response depending on the result
        if (result != null) {
            let wrappedResult: ResponseMetadata = result;
            if (!(wrappedResult instanceof ResponseMetadata)) {
                wrappedResult = new ResponseMetadata(200, wrappedResult);
            }

            callSendMethod(adapter, handler, wrappedResult, configuration.sendTypes[handlerName], adapterRequestData, next);
        }
    });
}

function unwrapAsyncValue<T>(promise: Promise<any>, cb: (_err: any, _value?: T) => void): void {
    Promise.resolve(promise)
        .done((unwrappedValue: any) => typeof unwrappedValue === "function" ? unwrappedValue(cb) : cb(null, unwrappedValue), cb);
}

function callSendMethod(adapter: ExpressAdapter, handler: Function, response: ResponseMetadata,
    sendType: SendType, adapterRequestData: any, next: Function): void {
    unwrapAsyncValue(response.body, (err: Error, body: any) => {
        if (err) {
            return next(err);
        }

        switch (sendType) {
            case SendType.JSON:
                adapter.sendJson(response.statusCode, body, adapterRequestData, response.headers);
                break;
            default:
                adapter.send(response.statusCode, body, adapterRequestData, response.headers);
                break;
        }
    });
}

/**
 * Combine the root and path strings into a single string.
 *
 * Can't use default path package, as the paths can contain parameters and regexes.
 *
 * @param  {string} root root path
 * @param  {string} path path
 * @return {string} combined path
 */
function createPathWithRoot(root: string, path: string): string {
    if (root && path) {
        if (root[root.length - 1] !== "/" && path[0] !== "/") {
            return root + "/" + path;
        } else if (root[root.length - 1] === "/" && path[0] === "/") {
            return root + path.slice(1);
        } else {
            return root + path;
        }
    } else if (root) {
        return root;
    } else {
        return path;
    }
}

/**
 * For each marked method, add a configuration to the adapter.
 *
 * @param {ExpressAdapter}   adapter         [description]
 * @param {IControllerClass} controllerClass [description]
 */
export function applyConfiguration(adapter: ExpressAdapter, controllerClass: IControllerClass): void {
    // Acquire the configuration
    const config: IControllerConfiguration = controllerClass.prototype.$$controllerConfiguration;
    if (config === undefined) {
        throw new Error("Supplied controller is not a decorated controller.");
    }

    // Create the instance
    // TODO: standard constructor parameters?
    let instance: any = new controllerClass();

    // Install all middlewares
    config.middlewares.forEach((middleware: IMiddleware) => adapter.addMiddleware(config, createPathWithRoot(config.root, middleware.path),
        instance, middleware.handlerName, instance[middleware.handlerName]));

    // Install all routes
    config.routes.forEach((route: IRoute) => adapter.addRoute(config, route.method, createPathWithRoot(config.root, route.path),
        instance, route.handlerName, instance[route.handlerName], config.routesMiddlewares[route.handlerName]));
}

/////////////////////////////////////////////////////////////////////
/// Response info
/////////////////////////////////////////////////////////////////////

class ResponseMetadata {
    public statusCode: number;
    public headers: IHeader[];
    public body: any;

    constructor(statusCode: number);
    constructor(body: any);
    constructor(statusCode: number, body: any);

    constructor(statusCodeOrBody: number | any, maybeBody?: any) {
        if (typeof statusCodeOrBody === "number") {
            [this.statusCode, this.body] = [statusCodeOrBody, maybeBody];
        } else {
            [this.statusCode, this.body] = [200, statusCodeOrBody];
        }
    }
}
